"""
SoccerAction
~~~~~~~~~~~~

SoccerAction is a Python package for objectively quantifying the impact of the
individual actions performed by soccer players using event stream data.

Full documentation is at <https://ml-kuleuven.github.io/socceraction/>.
:copyright: (c) 2020 by DTAI KU Leuven.
:license: MIT, see LICENSE for more details.
"""
__version__ = '1.4.2'
